@@ -10,12 +10,12 @@ value of corresponding (environment) variable.
 For each string value with ``$YP_TIME_*`` or ``${YP_TIME_*}`` syntax,
 substitute with value of corresponding date-time string.
 
-Validate against specified JSON schema if root file starts with a
-``#!<SCHEMA-URI>`` line.
-
+Validate against specified JSON schema if root file starts with either
+``#!<SCHEMA-URI>`` or ``# yaml-language-server: $schema=<SCHEMA-URI>`` line.
 """
 
 from argparse import ArgumentParser, RawDescriptionHelpFormatter
+from contextlib import suppress
 from datetime import datetime
 from errno import ENOENT
 import logging
@@ -31,8 +31,8 @@ from dateutil.relativedelta import relativedelta
 from dateutil.tz import tzlocal
 import jmespath
 import jsonschema
-import yaml
-from yaml.dumper import SafeDumper
+from ruamel.yaml import YAML
+from ruamel.yaml.constructor import ConstructorError
 
 from . import __version__
 
@@ -89,6 +89,8 @@ def strftime_with_colon_z(dto: datetime, time_format: str):
     """
     utcoffset = dto.utcoffset()
     if utcoffset is None:
+        for code in ('%z', '%:z', '%::z', '%:::z'):
+            time_format = time_format.replace(code, '')
         return dto.strftime(time_format)
     # Hopefully, we don't need to have sub-seconds in time zones.
     offset_total_seconds = int(utcoffset.total_seconds())
@@ -129,24 +131,26 @@ class UnboundVariableError(ValueError):
     __str__ = __repr__
 
 
-class YpSafeDumper(SafeDumper):
-    """Override dumping method for a time stamp to dump out str."""
-
-    @classmethod
-    def set_time_format(cls, time_format: str):
-        """Set time format."""
-        cls.time_format = time_format
-
-    def represent_datetime(self, data: datetime):
-        """Dump datetime as string."""
-        if hasattr(self, 'time_format'):
-            value = strftime_with_colon_z(data, self.time_format)
-            return self.represent_scalar('tag:yaml.org,2002:str', value)
-        else:
-            return super().represent_datetime(data)
+def construct_yaml_timestamp(constructor, node):
+    """Return a method to add to the YAML constructor to parse datetime."""
+    print(1, file=sys.stderr)
+    try:
+        return datetimeparse(node.value)
+    except ValueError:
+        raise ConstructorError(
+            None,
+            None,
+            f'failed to construct timestamp from "{node.value}"',
+            node.start_mark,
+        )
 
 
-YpSafeDumper.add_representer(datetime, YpSafeDumper.represent_datetime)
+def get_represent_datetime(time_format: str):
+    """Return a method to add to the YAML representer to represent datetime."""
+
+    return lambda representer, data: representer.represent_scalar(
+        'tag:yaml.org,2002:str',
+        strftime_with_colon_z(data, time_format))
 
 
 class DataProcessor:
@@ -167,6 +171,11 @@ class DataProcessor:
 
        Turn on/off variable substitution.
 
+    .. py:attribute:: .include_dict
+       :type: dict
+
+       Dictonary for values that can be substituted for include.
+
     .. py:attribute:: .include_paths
        :type: list
 
@@ -210,10 +219,41 @@ class DataProcessor:
        :type: str
 
        Value to substitute for unbound variables.
+
+    .. py:attribute:: .INCLUDE_SCHEMA
+       :type: dict
+
+       (Class) The schema of the INCLUDE syntax.
     """
 
     INCLUDE_KEY = 'INCLUDE'
+    MERGE_KEY = 'MERGE'
     QUERY_KEY = 'QUERY'
+    VARIABLES_KEY = 'VARIABLES'
+
+    INCLUDE_SCHEMA = {
+        'properties': {
+            INCLUDE_KEY: {
+                'type': 'string',
+            },
+            MERGE_KEY: {
+                'type': 'boolean',
+            },
+            QUERY_KEY: {
+                'type': 'string',
+            },
+            VARIABLES_KEY: {
+                'patternProperties': {
+                    r'^[A-z_]\w*$': {'type': 'string'},
+                },
+                'additionalProperties': False,
+                'type': 'object',
+            },
+        },
+        'additionalProperties': False,
+        'required': [INCLUDE_KEY],
+        'type': 'object',
+    }
 
     REC_SUBSTITUTE = re.compile(
         r"\A"
@@ -223,6 +263,7 @@ class DataProcessor:
         r"\$"
         r"(?P<brace_open>\{)?"
         r"(?P<name>[A-z_]\w*)"
+        r"(?P<cast>\.(?:int|float|bool))?"
         r"(?(brace_open)\})"
         r")"
         r"(?P<tail>.*)"
@@ -247,6 +288,7 @@ class DataProcessor:
             item
             for item in os.getenv('YP_INCLUDE_PATH', '').split(os.pathsep)
             if item)
+        self.include_dict = {}
         self.schema_prefix = os.getenv('YP_SCHEMA_PREFIX')
         self.time_formats = {'': '%FT%T%:z'}
         self.time_now = datetime.now(tzlocal())  # assume application is fast
@@ -267,38 +309,80 @@ class DataProcessor:
         in_filename = self.get_filename(in_filename, [])
         root = self.load_file(in_filename)
         schema_location = self.load_file_schema(in_filename)
-        stack = [[root, [in_filename]]]
+        stack = [[root, [in_filename], self.variable_map]]
         while stack:
-            data, parent_filenames = stack.pop()
-            data = self.process_variable(data)
-            data = self.load_include_file(data)[0]
+            data, parent_filenames, variable_map = stack.pop()
+            data = self.process_variable(data, variable_map)
+            if data is root:
+                # Handle INCLUDE at root level.
+                # Ignore the MERGE flag.
+                data, parent_filenames, variable_map = self.load_include_file(
+                    data, parent_filenames, variable_map)[0:3]
+            type_of_data = type(data)
             items_iter = None
-            if isinstance(data, list):
+            skip_keys = set()
+            if type_of_data is list:
                 items_iter = enumerate(data)
-            elif isinstance(data, dict):
-                items_iter = data.items()
+            elif type_of_data is dict:
+                items_iter = data.copy().items()
             if items_iter is None:
                 continue
             for key, item in items_iter:
-                item = data[key] = self.process_variable(item)
-                include_data, parent_filenames = self.load_include_file(
-                    item, parent_filenames)
-                if include_data != item:
+                if key in skip_keys:
+                    continue
+                item = data[key] = self.process_variable(item, variable_map)
+                include_data, parent_filenames_x, variable_map_x, is_merge = (
+                    self.load_include_file(
+                        item, parent_filenames, variable_map))
+                if is_merge and type_of_data != type(include_data):
+                    raise TypeError()
+                if is_merge and type_of_data is list:
+                    # For a list, the iterator seems to handle the new items
+                    # perfectly fine. We insert the included list at and after
+                    # the current position. The current item is logically
+                    # replaced by the first item of the inserted list.
+                    del data[key]
+                    item = None
+                    for i, include_item in enumerate(include_data):
+                        data.insert(key + i, include_item)
+                        if i == 0:
+                            item = include_item
+                elif is_merge and type_of_data is dict:
+                    # For a dict, the iterator cannot handle size changes, so
+                    # we can only iterate over a copy of the original dict. We
+                    # insert the items in the dict normally, but we'll need to
+                    # add elements of the dict to the stack to ensure we visit
+                    # any sub-trees for include files, etc.
+                    del data[key]
+                    item = None
+                    for include_key, include_item in include_data.items():
+                        data[include_key] = include_item
+                        skip_keys.add(include_key)
+                        if (
+                            isinstance(include_item, dict)
+                            or isinstance(include_item, list)
+                        ):
+                            stack.append([
+                                include_item,
+                                parent_filenames_x,
+                                variable_map_x,
+                            ])
+                elif include_data != item:
                     item = data[key] = include_data
                 if isinstance(item, dict) or isinstance(item, list):
-                    stack.append([data[key], parent_filenames])
+                    stack.append(
+                        [data[key], parent_filenames_x, variable_map_x])
         if out_filename == '-':
             out_file = sys.stdout
         else:
             out_file = open(out_filename, 'w')
-        YpSafeDumper.set_time_format(self.time_formats[''])
-        # Set sort_keys=False to preserve dict ordering (with Python 3.7+)
-        yaml.dump(
-            root,
-            out_file,
-            Dumper=YpSafeDumper,
-            default_flow_style=False,
-            sort_keys=False)
+        yaml = YAML(typ='safe', pure=True)
+        yaml.default_flow_style = False
+        yaml.sort_base_mapping_type_on_output = False
+        yaml.representer.add_representer(
+            datetime,
+            get_represent_datetime(self.time_formats['']))
+        yaml.dump(root, out_file)
         self.validate_data(root, out_filename, schema_location)
 
     def get_filename(self, filename: str, parent_filenames: list) -> str:
@@ -332,32 +416,53 @@ class DataProcessor:
     def load_include_file(
         self,
         value: object,
-        parent_filenames: list = None,
+        parent_filenames: list,
+        variable_map: dict,
     ) -> tuple:
-        """Load data if value indicates the root file or an include file.
+        """Load data if value indicates an include file.
 
         :param value: Value that may contain file name to load.
         :param parent_filenames: Stack of parent file names.
+        :param variable_map: :py:attr:`.variable_map` in the local scope,
+                             may have additional variables.
         """
-        if parent_filenames is None:
-            parent_filenames = []
-        else:
-            parent_filenames = list(parent_filenames)
-        while (
-            self.is_process_include
-            and isinstance(value, dict)
-            and self.INCLUDE_KEY in value
-        ):
+        orig_value = value
+        parent_filenames = list(parent_filenames)
+        variable_map = dict(variable_map)
+        is_merge = False
+        while self.is_process_include and self._is_include(value):
+            is_merge = (self.MERGE_KEY in orig_value)
             include_filename = self.process_variable(
                 value[self.INCLUDE_KEY])
-            filename = self.get_filename(include_filename, parent_filenames)
+            try:
+                loaded_value = self.include_dict[include_filename]
+                filename = include_filename
+            except KeyError:
+                filename = self.get_filename(
+                    include_filename, parent_filenames)
+                loaded_value = self.load_file(filename)
             parent_filenames.append(filename)
-            loaded_value = self.load_file(filename)
+            if self.VARIABLES_KEY in value:
+                variable_map.update(value[self.VARIABLES_KEY])
             if self.QUERY_KEY in value:
                 value = jmespath.search(value[self.QUERY_KEY], loaded_value)
             else:
                 value = loaded_value
-        return value, parent_filenames
+        return value, parent_filenames, variable_map, is_merge
+
+    @classmethod
+    def _is_include(cls, value: object) -> bool:
+        """Return True if value is recognised as an INCLUDE syntax.
+
+        :param value: Value that may contain file name to load.
+        :return: True if value is recognised as an INCLUDE syntax.
+        """
+        if not isinstance(value, dict) or cls.INCLUDE_KEY not in value:
+            return False
+        with suppress(jsonschema.exceptions.ValidationError):
+            jsonschema.validate(schema=cls.INCLUDE_SCHEMA, instance=value)
+            return True
+        return False
 
     @staticmethod
     def load_file(filename: str) -> object:
@@ -366,15 +471,19 @@ class DataProcessor:
         :param filename: name of file to load content.
         :return: the loaded data structure.
         """
+        yaml = YAML(typ='safe', pure=True)
+        yaml.constructor.add_constructor(
+            'tag:yaml.org,2002:timestamp',
+            construct_yaml_timestamp)
         if filename == '-':
-            return yaml.safe_load(sys.stdin)
+            return yaml.load(sys.stdin)
         else:
             with open(filename) as file_:
-                return yaml.safe_load(file_)
+                return yaml.load(file_)
 
     @staticmethod
     def load_file_schema(filename: str) -> object:
-        """Load schema location from #! line of file.
+        """Load schema location from the schema association line of file.
 
         :param filename: name of file to load schema location.
         :return: a string containing the location of the schema or None.
@@ -384,12 +493,17 @@ class DataProcessor:
         else:
             with open(filename) as file_:
                 line = file_.readline()
-        if line.startswith('#!'):
-            return line[2:].strip()
+        for prefix in ('#!', '# yaml-language-server: $schema='):
+            if line.startswith(prefix):
+                return line[len(prefix):].strip()
         else:
             return None
 
-    def process_variable(self, item: object) -> object:
+    def process_variable(
+        self,
+        item: object,
+        variable_map: dict = None,
+    ) -> object:
         """Substitute environment variables into a string value.
 
         Return `item` as-is if not `.is_process_variable` or if `item` is not a
@@ -407,11 +521,15 @@ class DataProcessor:
 
         :param item: Item to process. Do nothing if not a str.
         :return: Processed item on success.
+        :param variable_map: :py:attr:`.variable_map` in the local scope,
+                             may have additional variables.
 
         """
         if not self.is_process_variable or not isinstance(item, str):
             return item
-        ret = ""
+        if variable_map is None:
+            variable_map = self.variable_map
+        ret = ''
         try:
             tail = item.decode()
         except AttributeError:
@@ -420,25 +538,52 @@ class DataProcessor:
             match = self.REC_SUBSTITUTE.match(tail)
             if match:
                 groups = match.groupdict()
-                substitute = groups["symbol"]
-                if len(groups["escape"]) % 2 == 0:
-                    if groups["name"] in self.variable_map:
-                        substitute = self.variable_map[groups["name"]]
-                    elif groups["name"].startswith('YP_TIME'):
+                substitute = groups['symbol']
+                if len(groups['escape']) % 2 == 0:
+                    if groups['name'] in variable_map:
+                        substitute = variable_map[groups['name']]
+                    elif groups['name'].startswith('YP_TIME'):
                         substitute = self._process_time_variable(
-                            groups["name"])
+                            groups['name'])
                     elif self.unbound_placeholder is not None:
                         substitute = str(self.unbound_placeholder)
                     else:
-                        raise UnboundVariableError(groups["name"])
-                ret += (
-                    groups["head"]
-                    + groups["escape"][0:len(groups["escape"]) // 2]
-                    + substitute)
-                tail = groups["tail"]
+                        raise UnboundVariableError(groups['name'])
+                if groups['cast']:
+                    if groups['head'] or tail != item:
+                        raise ValueError(
+                            f'{item}: bad substitution expression')
+                    try:
+                        if groups['cast'] == '.int':
+                            substitute = int(substitute)
+                        elif groups['cast'] == '.float':
+                            substitute = float(substitute)
+                        elif (
+                            groups['cast'] == '.bool'
+                            and substitute.lower() in ('0', 'false', 'no')
+                        ):
+                            substitute = False
+                        elif (
+                            groups['cast'] == '.bool'
+                            and substitute.lower() in ('1', 'true', 'yes')
+                        ):
+                            substitute = True
+                        else:
+                            raise ValueError
+                    except ValueError:
+                        raise ValueError(
+                            f'{item}: bad substitution value: {substitute}')
+                    ret = substitute
+                    tail = ''
+                else:
+                    ret += (
+                        groups['head']
+                        + groups['escape'][0:len(groups['escape']) // 2]
+                        + substitute)
+                    tail = groups['tail']
             else:
                 ret += tail
-                tail = ""
+                tail = ''
         return ret
 
     def _process_time_variable(self, name: str) -> str:
